package net.java.dev.annsor;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;

/**
 * An AnnotationElement contains not only the annotation being processed, but also information
 * about the field, method, or constructor the annotation decorated. This object is passed into
 * the {@link Processor#process(AnnotationElement)} method to allow the {@link Processor} to
 * handle the request.
 * TODO: Should there be only one String getName() method and remove getMethodName, etc..
 * TODO: Should there be a ElementType getElementType() instead of isField(), etc..
 * @author Richard L. Burton III
 */
public interface AnnotationElement {

    /**
     * Sets a custom attribute on the AnnotationElement. This allows Processors
     * to pass data between each other by means of the AnnotationElement. It is
     * also used internally to maintain a list of {@link StatefulProcessor}s.
     * @param key   The alias of the value stored in the attribute.
     * @param value The value to store.
     */
    void setAttribute(String key, Object value);

    /**
     * This allows a {@link Processor} to remove a selected attribute from the
     * temporary store.
     * @param key The alias name of the object to remove.
     */
    void removeAttribute(String key);

    /**
     * Used to locate and return a vale stored in the attributes under a given
     * alias.
     * @param key The alias of the object to fetch.
     * @return The object stored under that location or null if not found.
     */
    Object getAttribute(String key);

    /**
     * This method checks to see if the provided scope is in range of the defined
     * scope attribute of the @Processor annotation.
     * @param type The scope to check.
     * @return True if the provided scope is in in scope, false otherwise.
     */
    boolean isScopeValid(Scope type);

    /**
     * Returns an array of parameter data types for the method decorated.
     * @return An array of Class if there's a method with arguements, null otherwise.
     */
    Class[] getMethodParameters();


    /**
     * Returns reference to the Attribute class type that the Annotation is assoicated with.
     * @return The attribute class type the Annotation is assoicated with.
     */
    Class getParameterType();

    /**
     * Returns the Annotation that is currently being processed.
     * @return The Annotation that is being processed.
     */
    Annotation getAnnotation();

    /**
     * Returns reference to the Method that the Annotation is assoicated with.
     * @return The method the Annotation is assoicated with.
     */
    Method getMethod();

    /**
     * Returns the method name.
     * @return The method name.
     */
    String getMethodName();

    /**
     * Returns reference to the Field that the Annotation is assoicated with.
     * @return The Field the Annotation is assoicated with.
     */
    Field getField();

    /**
     * Returns the field name.
     * @return The field name.
     */
    String getFieldName();

    /**
     * Returns the Class that the Annotation is referenced to.
     * @return The Class that the annotation is referenced to.
     */
    Class getTargetClass();

    /**
     * Returns the Class that the Annotation is referenced to.
     * @return Returns the Class that the Annotation is referenced to.
     */
    Constructor getConstructor();

    /**
     * Returns the Class name that the annotation is bound to.
     * @return Returns the Class name that the annotation is bound to.
     */
    String getTargetClassName();

    /**
     * The instance object that was decorated with the annotation.
     * @return The instance object that was decorated with the annotation.
     */
    Object getTarget();

    /**
     * Returns true if the Annotation is decorating a field.
     * @return True if the annotation is bound to a field, false otherwise.
     */
    boolean isField();

    /**
     * Returns true if the Annotation is decorating a method.
     * @return True if the annotation is bound to a method, false otherwise.
     */
    boolean isMethod();

    /**
     * Returns true if the Annotation is decorating a class.
     * @return True if the annotation is bound to a class, false otherwise.
     */
    boolean isClass();

    /**
     * Returns the Scope (FIELD, CLASS or METHOD)
     * @return The type of annotation.
     */
    Scope getScope();

    AnnsorContext getContext();

}
