package net.java.dev.annsor;

import net.java.dev.annsor.annotations.Factory;
import net.java.dev.annsor.descriptor.ClassDescriptor;
import net.java.dev.annsor.descriptor.ClassDescriptorBuilder;
import net.java.dev.annsor.processors.FactoryProcessor;
import net.java.dev.annsor.processors.PojoProcessor;
import net.java.dev.annsor.processors.PojoProcessorFactory;
import net.java.dev.annsor.processors.StatefulPojoProcessorFactory;
import static net.java.dev.annsor.util.AnnotationInspector.isPresentOnMethod;

import java.lang.annotation.Annotation;
import java.lang.ref.SoftReference;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Annsor is used to inspect objects at runtime and process annotations through a
 * subscription model. Processors subscribe to Annotations they would like to be informed
 * about without having the developer inspect the object in question. This class by design
 * is not thread safe when registering and deregistering Processors.
 * <p/>
 * @author Richard L. Burton III
 * @see Processor
 * @see net.java.dev.annsor.annotations.Processor
 * @see Processor
 * @see ProcessorFactory
 * @see StatefulProcessorFactoryImpl
 * @see Convertable
 * @see Scope
 * @see net.java.dev.annsor.AnnotationElement
 */
public class Annsor implements java.io.Serializable {


    /**
     * Contains all of the AnnotationProcessors that were registered.
     */
    private Map<Class<? extends Annotation>, List<Processor>> annotationHandlers = new HashMap<Class<? extends Annotation>, List<Processor>>();

    /**
     * This lookup map is used to improve perfomance in resolving the Annotation.class. There is a lot
     * of overhead with invoking getClass().getName() on the Annotation Proxy.
     */
    private ConcurrentHashMap<Class, Class> lookup = new ConcurrentHashMap<Class, Class>();

    /**
     *
     */
    private ConcurrentHashMap<Class, SoftReference<ClassDescriptor>> classDescriptorCache = new ConcurrentHashMap<Class, SoftReference<ClassDescriptor>>();

    private List<CacheableFilter> cacheableFilters = new ArrayList<CacheableFilter>();

    private Annsor parent;

    public Annsor() {
        cacheableFilters.add(new ProxyCacheableFilter());
    }

    /**
     * The invoke method will inspect the target object for annotations that decorated it. When an annotation
     * is found on a decorated element, the proper AnnnotationProcessor(s) will be called.
     * @param target The target object to invoke the annsor.
     */
    public void process(Object target) {
        process(target, new DefaultAnnsorContext());
    }

    /**
     * This method processes annotations on given class. Invoking this method will cause the
     * target object of the AnnotationElement to always be null.
     * @param clazz The class to inspect for annotations.
     */
    public void process(Class clazz) {
        process(clazz, new DefaultAnnsorContext());
    }

    public void process(Object target, AnnsorContext context) {
        if (target == null) {
            return;
        }

        processParent(target, context);
        Class clazz = target.getClass();
        ClassDescriptor descriptor = getClassDescriptor(clazz);
        InternalProcessor internalProcessor = new InternalProcessor(this);
        internalProcessor.process(target, descriptor, context);
    }

    public void process(Class clazz, AnnsorContext context) {
        if (clazz == null) {
            throw new NullPointerException("The provided class was null.");
        }

        processParent(clazz, context);
        ClassDescriptor descriptor = getClassDescriptor(clazz);
        InternalProcessor internalProcessor = new InternalProcessor(this);
        internalProcessor.process(null, descriptor, context);
    }

    /**
     * This method is used to register Processor classes. When the provided Annotation is found, the
     * Processor#invoke(AnnotationElement) method is invoked.
     * @param annotation The annotation to be notified about.
     * @param processors An array of Processor classes.
     * @throws NullPointerException If the Annotation or the Processor[] is null.
     */
    public void register(Class<? extends Annotation> annotation, Processor... processors) {

        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        if (processors == null) {
            throw new NullPointerException("A null reference for the Processor was passed in.");
        }


        List<Processor> handlerList = getProcessorList(annotation);
        if (handlerList == null) {
            handlerList = new ArrayList<Processor>();
            handlerList.addAll(Arrays.asList(processors));
            annotationHandlers.put(annotation, handlerList);
        } else {
            handlerList.addAll(Arrays.asList(processors));
        }

    }

    public void register(Class<? extends Annotation> annotation, ProcessorFactory... factories) {

        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        if (factories == null) {
            throw new NullPointerException("A null referencce for the ProcessorFactory(s) was passed in.");
        }

        for (ProcessorFactory factory : factories) {
            if (factory instanceof net.java.dev.annsor.StatefulProcessorFactory) {
                register(annotation, new StatefulProcessorFactoryImpl(factory));
            } else {
                register(annotation, new FactoryProcessor(factory));
            }

        }
    }

    /**
     * This method is used to register classes that have been decorated with the Processor annotation. When the
     * provided Annotation is found, all methods decorated with Processor will be invoked and passed all
     * of the information they require.
     * @param annotation The annotation to be notified about.
     * @param processors An array of classes that have been decorated with the Processor Annotation.
     * @throws NullPointerException If the Annotation or the Processor[] is null.
     */
    @SuppressWarnings("unchecked")
    public void register(Class<? extends Annotation> annotation, Object... processors) {

        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        if (processors == null) {
            throw new NullPointerException("A null reference for the Pojo Processor(s) was passed in.");
        }

        for (Object object : processors) {
            if (object instanceof ProcessorFactory) {
                register(annotation, (ProcessorFactory) object);
            } else {
                Class clazz = object.getClass();

                if (isPresentOnMethod(clazz, net.java.dev.annsor.annotations.Processor.class)) {
                    register(annotation, new PojoProcessor(object));
                }

                if (clazz.isAnnotationPresent(Factory.class)) {
                    Factory factory = (Factory) clazz.getAnnotation(Factory.class);
                    if (factory.stateful()) {
                        register(annotation, new StatefulProcessorFactoryImpl(new StatefulPojoProcessorFactory(object)));
                    } else {
                        register(annotation, new PojoProcessorFactory(object));
                    }
                }
            }
        }
    }

    /**
     * This method is used to remove a all classes registered for a given Annotation.
     * @param annotation The annoation to deregister all of the listners.
     * @return A list of registered Processor classes.
     */
    @SuppressWarnings("{unchecked}")
    public List<Processor> deregister(Class<? extends Annotation> annotation) {

        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        List<Processor> processors = annotationHandlers.remove(annotation);
        if (processors != null) {
            return processors;
        } else {
            return Collections.EMPTY_LIST;
        }
    }

    /**
     * This method checks to see if there are any registered handlers for a given Annotation.
     * @param annotation The annotation to check and see if any AnnotationProcessors were registered.
     * @return true if there is one or more AnnotationProcessors registered, false otherwise.
     */
    public boolean isRegister(Class<? extends Annotation> annotation) {

        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        return annotationHandlers.containsKey(annotation);
    }

    /**
     * Returns a list of AnnotationProcessors that were registered for a given Annotation.
     * @param annotation The annotation which to return the registered AnnotationProcessors.
     * @return An unmodifiable list of AnnotationProcessors. If no processors are found, an empty list is returned.
     */
    public List<Processor> getProcessors(Class<? extends Annotation> annotation) {

        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        List<Processor> processors = getProcessorList(annotation);
        if (processors != null) {
            return processors;
        }
        return Collections.emptyList();
    }

    /**
     * Sets the parent Annsor that will be called before
     * the current container.
     * @param parent The parent container.
     */
    public void setParent(Annsor parent) {
        if (parent != this) {
            this.parent = parent; // TODO: Check for circular references.
        } else {
            throw new IllegalArgumentException("The parent you're trying to set is the same instance of this container.");
        }
    }

    /**
     * Returns the parent container.
     * @return The parent Container.
     */
    public Annsor getParent() {
        return parent;
    }

    private void processParent(Class target, AnnsorContext context) {
        if (parent != null) {
            parent.process(target, context);
        }
    }

    private void processParent(Object target, AnnsorContext context) {
        if (parent != null) {
            parent.process(target, context);
        }
    }

    /**
     * Locates the ClassDescriptor for the provided class.
     * @param clazz The class to locate the ClassDescriptor for.
     * @return An instance of ClassDescriptor if one is found.
     */
    private ClassDescriptor getClassDescriptor(Class clazz) {

        if (clazz == null) {
            throw new NullPointerException("A null reference for the class was passed in.");
        }

        if (!isCacheable(clazz)) {
            ClassDescriptorBuilder builder = new ClassDescriptorBuilder(clazz);
            return builder.build();
        }

        SoftReference<ClassDescriptor> softRef;
        softRef = classDescriptorCache.get(clazz);

        if (softRef == null || softRef.get() == null) {
            ClassDescriptorBuilder builder = new ClassDescriptorBuilder(clazz);
            ClassDescriptor cd = builder.build();
            softRef = new SoftReference<ClassDescriptor>(cd);
            classDescriptorCache.put(clazz, softRef);
        }
        return softRef.get();
    }

    /**
     * Locates the processors for the given Annotation.
     * @param annotation The annotation to locate the processors.
     * @return A list of AnnotationProcessors registered for the Annotation.
     */
    protected List<Processor> getProcessorList(Annotation annotation) {
        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        Class clazz = getKey(annotation);
        return annotationHandlers.get(clazz); // TODO: If there are no handlers, Return an EmptyList
    }

    /**
     * Locates the processors for the given Annotation.
     * @param annotation The annotation to locate the processors.
     * @return A list of AnnotationProcessors registered for the Annotation.
     */
    protected List<Processor> getProcessorList(Class<? extends Annotation> annotation) {

        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        return annotationHandlers.get(annotation); // TODO: Empty List if not found.
    }

    /**
     * Adds a new CacheableFilter to the chain of filters.
     * @param filter The filter to add.
     */
    public void addCacheableFilter(CacheableFilter filter){
        cacheableFilters.add(filter);
    }

    /**
     * This method iterates over the CacheableFilters and checks to see
     * if the given class is cacheable or not.
     * @param clazz The class to check if its cacheable.
     * @return true if it is, false otherwise.
     */
    protected boolean isCacheable(Class clazz){
        int size = cacheableFilters.size();
        for(int i=0;i<size;i++){
            CacheableFilter filter = cacheableFilters.get(i);
            if(!filter.isCacheable(clazz)){
                return false;
            }
        }
        return true;
    }

    protected Class getKey(Annotation annotation) {

        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        Class clazz = lookup.get(annotation.getClass());
        if (clazz == null) {
            clazz = getClass(annotation);
            lookup.putIfAbsent(annotation.getClass(), clazz);
        }
        return clazz;
    }

    private Class getClass(Annotation annotation) {

        if (annotation == null) {
            throw new NullPointerException("A null reference for the Annotation class was passed in.");
        }

        Class[] interfaces = annotation.getClass().getInterfaces();
        for (Class clazz : interfaces) {
            if (clazz != Annotation.class) {
                return clazz;
            }
        }
        throw new RuntimeException();
    }

}