package net.java.dev.annsor;

import static net.java.dev.annsor.Scope.*;
import net.java.dev.annsor.processors.PojoProcessor;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

/**
 * The default AnnotationElement to be passed around. The interface is the only
 * known contract to the AnnotationProcessors so the setter methods should never be
 * invoked. They only exist for performance reasons.
 * <p/>
 * @author Richard L. Burton III
 * @version 1.0
 */
class DefaultAnnotationElement implements AnnotationElement {

    private static final Logger LOGGER = Logger.getLogger(DefaultAnnotationElement.class.getName());

    private Map<String, Object> attributes;

    /**
     * The <code>Annotation</code> that decorated the class, field or method.
     */
    private Annotation annotation;

    /**
     * The <code>Method</code> that the annotation decorated.
     */
    private Method method;

    /**
     * The <code>Field</code> that the annotation decorated.
     */
    private Field field;

    /**
     * The <code>Class</code> that the annotation decorated.
     */
    private Class targetClass;

    /**
     * The target object that was decorated with the annotations.
     */
    private final Object target;

    /**
     * The Scope for the AnnotationElement.
     */
    private Scope scope;

    /**
     * The constructor that the annotation decorated.
     */
    private Constructor ctor;

    private Class parameterType;

    private Class[] methodParameters;

    private Map<String, Object> statefulProcessors;

    private Map<String, Object> statefulProcessorFactory;

    /** The AnnsorContext used to pass in values and obtain values from the processing.*/
    private AnnsorContext context;

    /**
     * Construts an instance of the AnnotationElemen that defaults to the a Class
     * Scope.
     * @param target      The target object that was decorated.
     * @param targetClass The target class.
     */
    DefaultAnnotationElement(Object target, Class targetClass, AnnsorContext context) {
        this.target = target;
        this.context = context;
        setTargetClass(targetClass);
    }

    public void addStatefulProcessor(String key, Object processor) {
        if (statefulProcessors == null) {
            statefulProcessors = new HashMap<String, Object>();
        }
        if(processor instanceof Processor){
            statefulProcessors.put(key, processor);
        }else{
            Processor pojoProcessor =  new PojoProcessor(processor);
            statefulProcessors.put(key, pojoProcessor);
        }
    }

    public Object getStatefulProcessor(String key) {
        if (statefulProcessors != null) {
            return statefulProcessors.get(key);
        } else {
            return null;
        }

    }

    public void addStatefulProcessorFactory(String key, Object processor) {
        if (statefulProcessorFactory == null) {
            statefulProcessorFactory = new HashMap<String, Object>();
        }
        statefulProcessorFactory.put(key, processor);
    }

    public Object getStatefulProcessorFactory(String key) {
        if (statefulProcessorFactory != null) {
            return statefulProcessorFactory.get(key);
        } else {
            return null;
        }

    }

    public void destroy() {

        if (statefulProcessorFactory != null) {
            for (Object processor : statefulProcessorFactory.values()) {
                try {
                    ((StatefulProcessorFactory) processor).destroy(this);
                } catch (Exception e) {
                    LOGGER.finer(e.getMessage());
                }
            }
        }
        if (statefulProcessors != null) {
            for (Object processor : statefulProcessors.values()) {
                try {
                    if(processor instanceof StatefulProcessor){
                        ((StatefulProcessor) processor).destroy(this);
                    }else{
                        ((PojoProcessor)processor).destroy(this);
                    }
                } catch (Exception e) {
                    LOGGER.finer(e.getMessage());
                }
            }
        }
    }

    public Class[] getMethodParameters() {
        return methodParameters;
    }

    public void setMethodParameters(Class[] methodParameters) {
        this.methodParameters = methodParameters;
    }

    public Annotation getAnnotation() {
        return annotation;
    }

    public Method getMethod() {
        return method;
    }

    /**
     * Returns the method name.
     * @return The method name.
     */
    public String getMethodName() {
        if (method != null) {
            return method.getName();
        } else {
            return null;
        }
    }

    public Field getField() {
        return field;
    }

    /**
     * Returns the field name.
     * @return The field name.
     */
    public String getFieldName() {
        if (field != null) {
            return field.getName();
        } else {
            return null;
        }
    }

    public Class getTargetClass() {
        return targetClass;
    }

    public Constructor getConstructor() {
        return ctor;
    }

    /**
     * Returns the Class name that the annotation is bound to.
     * @return Returns the Class name that the annotation is bound to.
     */
    public String getTargetClassName() {
        if (targetClass != null) {
            return targetClass.getName();
        } else {
            return null;
        }
    }

    public Object getTarget() {
        return target;
    }

    public boolean isField() {
        return field != null;
    }

    public boolean isMethod() {
        return method != null;
    }

    public boolean isClass() {
        return targetClass != null;
    }

    public Scope getScope() {
        return scope;
    }

    public AnnsorContext getContext() {
        return context;
    }

    void setAnnotation(Annotation annotation) {
        this.annotation = annotation;
    }

    void setMethod(Method method) {
        this.method = method;
        this.scope = METHOD;
    }

    void setField(Field field) {
        this.field = field;
        this.scope = FIELD;
    }

    void setTargetClass(Class targetClass) {
        this.targetClass = targetClass;
        this.scope = CLASS;
    }

    void reset() {
        this.targetClass = null;
        this.method = null;
        this.field = null;
        this.scope = null;
    }

    void setConstructor(Constructor ctor) {
        this.ctor = ctor;
        this.scope = CONSTRUCTOR;
    }

    public Class getParameterType() {
        return parameterType;
    }

    public void setParameterType(Class parameterType) {
        this.parameterType = parameterType;
        this.scope = PARAMETER;
    }

    public void setAttribute(String key, Object value) {
        getAttributes().put(key, value);
    }

    public void removeAttribute(String key) {
        attributes = getAttributes();
        Object value = attributes.get(key);
        attributes.remove(value);
    }

    public Object getAttribute(String key) {
        return getAttributes().get(key);
    }

    private Map<String, Object> getAttributes() {
        if (attributes == null) {
            attributes = new HashMap<String, Object>();
        }
        return attributes;
    }

    public boolean isScopeValid(Scope type) {
        return type == scope;
    }

    public String toString() {
        return "DefaultAnnotationElement{" +
                "annotation=" + annotation +
                ", attributes=" + attributes +
                ", method=" + method +
                ", field=" + field +
                ", targetClass=" + targetClass +
                ", target=" + target +
                ", scope=" + scope +
                ", ctor=" + ctor +
                ", parameterType=" + parameterType +
                ", methodParameters=" + (methodParameters == null ? null : Arrays.asList(methodParameters)) +
                '}';
    }

}