package net.java.dev.annsor;

import net.java.dev.annsor.descriptor.AttributeDescriptor;
import net.java.dev.annsor.descriptor.MethodDescriptor;
import net.java.dev.annsor.descriptor.ClassDescriptor;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.util.List;

/**
 * This class is designed to iterate over the elements of an object that have been decorated with
 * annotations and invoke the registered Processor.
 * @author Richard L. Burton III
 */
class InternalProcessor {

    /**
     * The Annsor container that contains the registeration of the Processors as well as
     * the descriptors of the classes.
     */
    private final Annsor container;

    /**
     * Constructs an instance of the object which reference to the owning Annsor container.
     * @param container The owning Annsor.
     */
    public InternalProcessor(Annsor container) {
        this.container = container;
    }

    /**
     * The main entry point for processing the target instance object. This causes a processing
     * of all methods, constructors, fields, and class level annotations.
     * @param target     The instance object that the annotation was found on.
     * @param descriptor A cached description of the class.
     */
    public void process(Object target, ClassDescriptor descriptor, AnnsorContext context) {
        DefaultAnnotationElement annotationElement = new DefaultAnnotationElement(target, descriptor.getAttribute(), context);
        processClassAnnotations(annotationElement, descriptor);
        processConstructors(annotationElement, descriptor);
        annotationElement.reset();
        processFields(annotationElement, descriptor);
        annotationElement.reset();
        processMethods(annotationElement, descriptor);
        annotationElement.destroy();
    }

    /**
     * This method iterates over all of the class level annotations and invokes the processors that were
     * decorated at the class level.
     * @param annotationElement The AnnotationElement with all of the meta data about the current Annotation.
     * @param descriptor        The ClassDescriptor to be processed for annotations.
     */
    void processClassAnnotations(DefaultAnnotationElement annotationElement, ClassDescriptor descriptor) {
        final Annotation[] annotations = descriptor.getAnnotations();
        for (int a = 0; a < annotations.length; a++) {
            Annotation annotation = annotations[a];
            annotationElement.setAnnotation(annotation);
            executeProcessors(annotationElement);
        }
    }

    /**
     * This method iterates over the annotations found at the constructor level of the class and invokes the
     * Processors registered for each annotation.
     * @param annotationElement The AnnotationElement with all of the meta data about the current Annotation.
     * @param descriptor        The class to be processed for annotations.
     */
    void processConstructors(DefaultAnnotationElement annotationElement, ClassDescriptor descriptor) {
        final List<AttributeDescriptor<Constructor>> ctors = descriptor.getConstructors();

        for (int c = 0; c < ctors.size(); c++) {

            final AttributeDescriptor<Constructor> ctor = ctors.get(c);
            final Annotation[] annotations = ctor.getAnnotations();
            annotationElement.setConstructor(ctor.getAttribute());

            for (int a = 0; a < annotations.length; a++) {
                Annotation annotation = annotations[a];
                annotationElement.setAnnotation(annotation);
                annotationElement.setConstructor(ctor.getAttribute());
                executeProcessors(annotationElement);
            }
        }
    }

    /**
     * This method iterates over each field located on each field and invokes the registered processes.
     * @param annotationElement The AnnotationElement with all of the meta data about the current Annotation.
     * @param descriptor        The class to be processed for annotations.
     */
    void processFields(DefaultAnnotationElement annotationElement, ClassDescriptor descriptor) {
        List<AttributeDescriptor<Field>> fields = descriptor.getFields();

        for (int f = 0; f < fields.size(); f++) {

            AttributeDescriptor<Field> field = fields.get(f);
            Annotation[] annotations = field.getAnnotations();
            annotationElement.setField(field.getAttribute());

            for (int a = 0; a < annotations.length; a++) {
                Annotation annotation = annotations[a];
                annotationElement.setAnnotation(annotation);
                executeProcessors(annotationElement);
            }
        }
    }

    /**
     * This method iterates over the method level annotations and invokes the registered Processors
     * @param annotationElement The AnnotationElement with all of the meta data about the current Annotation.
     * @param descriptor        The class to be processed for annotations.
     */
    void processMethods(DefaultAnnotationElement annotationElement, ClassDescriptor descriptor) {
        if (!descriptor.isMethodPresent())
            return;

        final List<MethodDescriptor> methods = descriptor.getMethods();

        for (int m = 0; m < methods.size(); m++) {

            final MethodDescriptor method = methods.get(m);
            annotationElement.setMethod(method.getAttribute());

            if (method.isParametersPresent())
                processMethodParameters(annotationElement, method);

            if (method.hasAnnotations()) {

                final Annotation[] annotations = method.getAnnotations();
                annotationElement.setMethodParameters(method.getParameterTypes());

                for (int a = 0; a < annotations.length; a++) {
                    final Annotation annotation = annotations[a];
                    annotationElement.setAnnotation(annotation);
                    executeProcessors(annotationElement);
                }
            }
        }
    }

    /**
     * This method iterates over the method argument level annotations and invokes the registered Processes.
     * @param annotationElement The AnnotationElement with all of the meta data about the current Annotation.
     * @param method            The class to be processed for annotations.
     */
    void processMethodParameters(DefaultAnnotationElement annotationElement, MethodDescriptor method) {
        final List<AttributeDescriptor<Class>> parameters = method.getParameters();
        annotationElement.setMethodParameters(method.getParameterTypes());

        for (int p = 0; p < parameters.size(); p++) {

            final AttributeDescriptor<Class> parameter = parameters.get(p);
            final Annotation[] annotations = parameter.getAnnotations();
            annotationElement.setParameterType(parameter.getAttribute());

            for (int a = 0; a < annotations.length; a++) {
                final Annotation annotation = annotations[a];
                annotationElement.setAnnotation(annotation);
                executeProcessors(annotationElement);
            }
        }
        annotationElement.setMethod(method.getAttribute());
    }

    /**
     * This method handles invoking each of the register processors for a given annotation.
     * @param element The AnnotationElement with all of the meta data about the current Annotation.
     */
    void executeProcessors(AnnotationElement element) {
        final List<Processor> handlers = container.getProcessorList(element.getAnnotation());

        if (handlers != null) {
            for (int h = 0; h < handlers.size(); h++) {
                final Processor handler = handlers.get(h);
                try {
                    handler.process(element);
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }
            }
        }
    }

}

