package net.java.dev.annsor.annotations;

import net.java.dev.annsor.Scope;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * This Annotation is used to flag a method has being an Annotation processor. The signature
 * of the method will be inspected to determine what values should be passed in during the
 * invocation.
 * <p/>
 * Parameter Type       Value
 * --------------       ----------
 * Field                The field that was decorated.
 * Method               The method that was decorated
 * Constructor          The constructor that was decorated.
 * Class                The class of the object being processed.
 * Class[]              The parameters to the Method or Constructor decorated.
 * String               The attribute name that was decorated.
 * AnnotationElement    An instance of AnnotationElement
 * Object               The instance of the object being processed. The type can be narrowed
 * down to the extact type. E.g., Person instead of Object
 * Annotation           The annotation decorated on the attribute.
 * @author Richard L. Burton III
 * @version 1.0
 */
@Target(ElementType.METHOD)
@Retention(RetentionPolicy.RUNTIME)
public @interface Processor {

    /**
     * The scope in which the processor method should be called on.
     * @return The scope in which the processor method should be invoked on.
     */
    Scope[] scope() default Scope.ANY;

    /**
     * Sets the index of the instance object being inspected.
     * @return The index of the target object to be passed into the method.
     */
    int methodParametersIndex() default -1;


    /**
     * Sets the index of the instance object being inspected.
     * @return The index of the target object to be passed into the method.
     */
    int targetObjectIndex() default -1;

    /**
     * Sets the index of the annotation object being inspected.
     * @return The index of the annotation to be passed into the method.
     */
    int annotationIndex() default -1;

    /**
     * Sets the index of the method object being inspected.
     * @return The index of the method to be passed into the method.
     */
    int methodIndex() default -1;

    /**
     * Sets the index of the field object being inspected.
     * @return The index of the field to be passed into the method.
     */
    int fieldIndex() default -1;

    /**
     * Sets the index of the field name object being inspected.
     * @return The index of the field name to be passed into the method.
     */
    int fieldNameIndex() default -1;

    /**
     * Sets the index of the Scope object being inspected.
     * @return The index of the Scope to be passed into the method.
     */
    int annotationTypeIndex() default -1;

    /**
     * Sets the index of the method name object being inspected.
     * @return The index of the method name to be passed into the method.
     */
    int methodNameIndex() default -1;

    /**
     * Sets the index of the target class  object being inspected.
     * @return The index of the target class to be passed into the method.
     */
    int targetClassIndex() default -1;

    /**
     * Sets the index of the target class name object being inspected.
     * @return The index of the target class name to be passed into the method.
     */
    int targetClassNameIndex() default -1;
          
    int annsorContextIndex() default -1;

}
