package net.java.dev.annsor.descriptor;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.util.LinkedList;

/**
 * A ClassDescriptor stores information about the decorated elements of the class.
 * @author Richard L. Burton III
 */
public class ClassDescriptor extends AttributeDescriptor<Class> {

    /**
     * The cache of constructors that has annotations assoicated with them.
     */
    private List<AttributeDescriptor<Constructor>> constructorCache = new ArrayList<AttributeDescriptor<Constructor>>();

    /**
     * The cache of methods that has annotations assoicated with them.
     */
    private List<MethodDescriptor> methodCache = new ArrayList<MethodDescriptor>();

    /**
     * The cache of fields that has annotations assoicated with them.
     */
    private List<AttributeDescriptor<Field>> fieldCache = new ArrayList<AttributeDescriptor<Field>>();

    private boolean fieldsPresent;

    private boolean methodPresent;

    private boolean parametersPresent;

    private Class sourceClass;

    /**
     * Constructs a ClassDescriptor using the provided class to build the internal cache.
     * @param clazz The class to inspect.
     */
    protected ClassDescriptor(Class clazz) {
        super(clazz);
        addField(clazz.getDeclaredFields());
        addMethod(clazz.getDeclaredMethods());
        addConstructor(clazz.getConstructors());
        this.sourceClass = clazz;
        fieldsPresent = fieldCache.size() != 0;
        methodPresent = methodCache.size() != 0;
    }

    /**
     * Adds the Constructors to the cache only if there are annotations decorating it.
     * @param constructors A list of Constructors to add.
     */
    protected void addConstructor(Constructor... constructors) {
        for (Constructor ctor : constructors) {
            if (ctor.getDeclaredAnnotations().length != 0)
                constructorCache.add(new AttributeDescriptor<Constructor>(ctor));
        }
    }

    protected void addConstructors(List<AttributeDescriptor<Constructor>> ctors){
        this.constructorCache.addAll(ctors);
    }

    /**
     * Adds the fields to the cache only if there are annotations decorating it.
     * @param fields A list of Fileds to add.
     */
    protected void addField(Field... fields) {
        for (Field field : fields) {
            if (field.getDeclaredAnnotations().length != 0)
                fieldCache.add(new AttributeDescriptor<Field>(field));
        }
    }

    /**
     * Adds the methods to the cache only if there are annotations decorating it.
     * @param methods A list of methods to add.
     */
    protected void addMethod(Method... methods) {
        for (Method method : methods) {
            MethodDescriptor methodDescriptor = new MethodDescriptor(method);
            if (methodDescriptor.hasAnnotations() || methodDescriptor.isParameterAnnotationsPresent()) {
                methodCache.add(new MethodDescriptor(method));
                if (!parametersPresent)
                    parametersPresent = methodDescriptor.isParametersPresent();
            }
        }
    }

    /**
     * Returns an unmodifiable list of method objects that have annotations
     * assoicated with them.
     * @return An unmodifiable list of method objects.
     */
    public List<MethodDescriptor> getMethods() {
        return methodCache;
    }

    // TODO: Is this right?
    // TODO: Should I add List<AttributeDescriptor> getAttributeElementsWithAnnotation(Class<? extends Annotation> annotation);
    // annsor.process(object,  context, In.class);
    // annsor.process(object, Begin.class, In.class, Out.class, End.class);
    public List<MethodDescriptor> getMethodsWithAnnotation(Class<? extends Annotation> annotation){
        List<MethodDescriptor> methodDescriptorList = new LinkedList<MethodDescriptor>();
        for(int i=0;i<methodCache.size();i++){
            MethodDescriptor method = methodCache.get(i);
            if(method.getAttribute().isAnnotationPresent(annotation)){
                methodDescriptorList.add(method);
            }
        }
        return methodDescriptorList;
    }

    protected void addMethods(List<MethodDescriptor> methods) {
        this.methodCache.addAll(methods);
    }

    /**
     * Returns an unmodifiable list of field objects that have annotations
     * assoicated with them.
     * @return An unmodifiable list of Field objects.
     */
    public List<AttributeDescriptor<Field>> getFields() {
        return fieldCache;
    }

    protected void addFields(List<AttributeDescriptor<Field>> fields){
        this.fieldCache.addAll(fields);
    }

    /**
     * Returns an unmodifiable list of Constructors that have annotations
     * assoicated with them.
     * @return An unmodifiable list of Constructors.
     */
    public List<AttributeDescriptor<Constructor>> getConstructors() {
        return constructorCache;
    }

    public boolean isFieldsPresent() {
        return fieldsPresent;
    }

    public boolean isMethodPresent() {
        return methodPresent && parametersPresent;
    }

    public Class getSourceClass() {
        return sourceClass;
    }

    protected ClassDescriptor lock(){
        constructorCache = Collections.unmodifiableList(constructorCache);
        fieldCache = Collections.unmodifiableList(fieldCache);
        methodCache = Collections.unmodifiableList(methodCache);
        for(MethodDescriptor md : methodCache){
            md.lock();
        }
        return this;
    }

    public String toString() {
        return "ClassDescriptor{" +
                "constructorCache=" + constructorCache +
                ", methodCache=" + methodCache +
                ", fieldCache=" + fieldCache +
                ", fieldsPresent=" + fieldsPresent +
                ", methodPresent=" + methodPresent +
                ", parametersPresent=" + parametersPresent +
                ", sourceClass=" + sourceClass +
                '}';
    }

}
