package net.java.dev.annsor.descriptor;

import java.util.Stack;

/**
 * A <tt>ClassDescriptorBuilder</tt> is designed to build an instance of <code>ClassDescriptor</class>
 * that describes the class that was provided. An important note about the returned ClassDescriptor is
 * that it contains information about the parent class as well. The decorated elements on the child are
 * merged with the ClassDescriptor of the parent.
 * @author Richard L. Burton III
 */
public class ClassDescriptorBuilder {

    /**
     * The parent ClassDescriptor that will be merged with the child ClassDescriptors.
     */
    private ClassDescriptor parent;

    /**
     * The stack of ClassDescriptors that need to be merged.
     */
    private Stack<ClassDescriptor> descriptors = new Stack<ClassDescriptor>();

    /**
     * Creates an instance of ClassDescriptorBuilder using the provided Class as the
     * source of the processing.
     * @param clazz The class to be inspected.
     */
    public ClassDescriptorBuilder(Class clazz) {
        ClassDescriptor leaf = new ClassDescriptor(clazz);
        descriptors.push(leaf);
        clazz = leaf.getSourceClass();
        while ((clazz = clazz.getSuperclass()) != Object.class) {
            descriptors.push(new ClassDescriptor(clazz));
        }
        parent = descriptors.pop();
    }

    /**
     * Inspects the Stack of ClassDescriptors and merges the child class with the parent.
     * @return A ClassDescriptor of the provided Class.
     */
    public ClassDescriptor build() {
        while (!descriptors.isEmpty()) {
            merge(descriptors.pop());
        }
        return parent.lock();
    }

    /**
     * Merges the decorated elements of the child with the parent.
     * @param cd The child ClassDescriptor to merge with the parent.
     */
    private void merge(ClassDescriptor cd) {
        parent.addConstructors(cd.getConstructors());
        parent.addFields(cd.getFields());
        parent.addMethods(cd.getMethods());

        if (cd.hasAnnotations()) {
            parent.addAnnotations(cd.getAnnotations());
        }

    }

    /**
     * Returns the ClassDescriptor that was generated from the
     * Class provided.
     * @return The ClassDescriptor of the class.
     */
    public ClassDescriptor getClassDescriptor() {
        return parent;
    }

}
