package net.java.dev.annsor.descriptor;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Collections;

/**
 * This class defines the Descriptor for decorated <code>Method</code>s on the class. An instance
 * of this class created under two conditions:
 * <br/>
 * a) The method has an annotation decorating it.<br/>
 * b) A parameter of the method is decorated with an Annotation.
 * <br/>
 * TODO: Handle exposing parameterTypes without having to clone the array.
 * @author Richard L. Burton III
 */
public class MethodDescriptor extends AttributeDescriptor<Method> {

    private List<AttributeDescriptor<Class>> parameters = new ArrayList<AttributeDescriptor<Class>>();

    private Class[] parameterTypes;

    private boolean parametersPresent;

    private boolean parameterAnnotationsPresent;

    public MethodDescriptor(Method attribute) {
        super(attribute);
        parameterTypes = attribute.getParameterTypes();
        Annotation[][] a = attribute.getParameterAnnotations();
        for (int i = 0; i < a.length; i++) {
            Annotation[] annotation = a[i];
            Class clazz = parameterTypes[i];
            AttributeDescriptor<Class> parameterDescriptor = new AttributeDescriptor<Class>(clazz, annotation);
            parameters.add(parameterDescriptor);
            if (!parameterAnnotationsPresent && parameterDescriptor.hasAnnotations()) {
                parameterAnnotationsPresent = true;
            }
        }

        parametersPresent = parameters.size() != 0;
    }

    public List<AttributeDescriptor<Class>> getParameters() {
        return parameters;
    }

    public Class[] getParameterTypes() {
        return parameterTypes;
    }


    public boolean isParametersPresent() {
        return parametersPresent;
    }

    public boolean isParameterAnnotationsPresent() {
        return parameterAnnotationsPresent;
    }

    public String toString() {
        return "MethodDescriptor{" +
                "parameterAnnotationsPresent=" + parameterAnnotationsPresent +
                ", parameters=" + parameters +
                ", parameterTypes=" + (parameterTypes == null ? null : Arrays.asList(parameterTypes)) +
                ", parametersPresent=" + parametersPresent +
                '}';
    }

    protected void lock() {
        parameters = Collections.unmodifiableList(parameters);
    }

}
