package net.java.dev.annsor.processors;

import net.java.dev.annsor.AnnotationElement;
import net.java.dev.annsor.Processor;
import net.java.dev.annsor.ProcessorFactory;

import java.lang.annotation.Annotation;
import java.util.HashMap;
import java.util.Map;

/**
 * The only real information that should be cached in this instance, would be the data that decorated the
 * Object Processor that was returned from the Factory.
 * @author Richard L. Burton III
 * @version 1.0
 */
public class FactoryProcessor implements Processor {

    private ProcessorFactory factory;

    private final Map<Annotation, PojoProcessor> cachedPojoProcessors = new HashMap<Annotation, PojoProcessor>();

    public FactoryProcessor(ProcessorFactory factory) {
        this.factory = factory;
    }

    /**
     * When a Annotation is found and a matching <code>Processor</code> class is registered,
     * the invoke(AnnotationElement) method called. The AnnotationElement provides insight to
     * the the artifacts that the Annotation decorates.
     * @param annotationElement A value object with all the required information about the Annotation found.
     * @throws Exception Thrown if there's a problem with a subsystem.
     */
    public void process(AnnotationElement annotationElement) throws Exception {
        Object handler = factory.create(annotationElement);
        if (handler instanceof Processor) {
            ((Processor) handler).process(annotationElement);
        } else {
            PojoProcessor processor = getPojoProcessor(annotationElement.getAnnotation(), handler);
            processor.process(annotationElement);
        }
    }

    private PojoProcessor getPojoProcessor(Annotation annotationClass, Object handler) {
        PojoProcessor processor;
        synchronized (cachedPojoProcessors) {
            processor = cachedPojoProcessors.get(annotationClass);
            if (processor == null) {
                processor = new PojoProcessor(handler);
                cachedPojoProcessors.put(annotationClass, processor);
            }
        }
        return processor;
    }

}
