package net.java.dev.annsor.processors;

import net.java.dev.annsor.AnnotationElement;
import net.java.dev.annsor.Convertable;
import net.java.dev.annsor.Scope;
import net.java.dev.annsor.AnnsorContext;
import net.java.dev.annsor.annotations.Convertor;
import net.java.dev.annsor.annotations.Create;
import net.java.dev.annsor.annotations.Destroy;
import net.java.dev.annsor.annotations.Processor;
import static net.java.dev.annsor.processors.MethodArgumentType.*;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

/**
 * This class is used to invoke a method that was decorated as being a Processor method. When a
 * method is decorated as being a processor, the method can either be inspected to see what arguments
 * should be provided or the index could be assigned using the Processor.
 * <p/>
 * TODO: StatefulProcessorFactory pojo objects Must accept AnnotationElement.
 * TODO: Better validate the method parameter types. e.g.. method(Annotation1 a, Annotation2 b)
 * a. Method invocation on the Annotation
 * @author Richard L. Burton III
 * @version 1.0
 */
class PojoMethod {

    /**
     * The class level logger.
     */
    private Logger LOGGER = Logger.getLogger(PojoMethod.class.getName());

    /**
     * The method decorated with the Processor annotation.
     */
    private Method processorMethod;

    /**
     * The method decorated with the Convertor annotation.
     */
    private Method convertableMethod;

    /**
     * Used to created the Object[] that will be passed into the Processor method.
     */
    private int argumentSize = 0;

    /**
     * The instance of the Annotation that decorated the processor method.
     */
    private PojoMethodParameters pojoMethodParameters;

    /**
     * A boolean flag that indicated that the handler contains a convertable method.
     */
    private boolean convertable = false;

    private MethodArgumentType[] processorParameterTypes;

    private MethodArgumentType[] convertorParameterTypes;

    private boolean[] indexValid = new boolean[8];

    private Class<? extends Annotation> annotation;

    private Class[] parameterClassTypes;

    /**
     * Constructs an instance of PojoMethod and caches the information about the
     * give methods in question.
     * @param processor The annotation Processor that was set.
     * @param handler   The Pojo that will be the AccessorProcessor.
     * @param method    The processorMethod decorated with the Processor annotation.
     */
    public PojoMethod(Processor processor, Object handler, Method method) {
        this(new PojoMethodParameters(processor), handler, method);
    }

    public PojoMethod(Create processor, Object handler, Method method) {
        this(new PojoMethodParameters(processor), handler, method);
    }

    public PojoMethod(Destroy processor, Object handler, Method method) {
        this(new PojoMethodParameters(processor), handler, method);
    }

    /**
     * @param parameters The parameters to invoke the method with.
     * @param handler    The owning class.
     * @param method     The method to invoke.
     * @throws IllegalArgumentException If the method is decorated has being a Convertor, but
     *                                  doesn't have a return type.
     */
    private PojoMethod(PojoMethodParameters parameters, Object handler, Method method) {
        this.pojoMethodParameters = parameters;
        convertable = handler instanceof Convertable;

        this.processorMethod = method;
        method.setAccessible(true);
        calculateArgumentsLength();
        parameterClassTypes = method.getParameterTypes();

        if (!convertable) {
            Method[] methods = handler.getClass().getDeclaredMethods();
            for (Method m : methods) {
                if (m.isAnnotationPresent(Convertor.class)) {
                    if (convertableMethod == null) {
                        if (m.getReturnType() == null) {
                            throw new IllegalArgumentException("The decorated @Convertor method " + m.getName() +
                                    " does not have a return type.");
                        }
                        convertableMethod = m;
                        convertableMethod.setAccessible(true);
                        convertorParameterTypes = getMethodParameterTypes(convertableMethod);
                        convertable = true;
                    } else {
                        LOGGER.info("Already found @Convertable on method " + convertableMethod.toString());
                    }
                }
            }
            processorParameterTypes = getMethodParameterTypes(method);
        }
    }

    /**
     * This method calculates the number of parameters requested by
     * the Processs arguments.
     */
    private void calculateArgumentsLength() {
        if (pojoMethodParameters.isTargetObjectIndex()) {
            argumentSize++;
            indexValid[0] = true;
        }
        if (pojoMethodParameters.isMethodIndex()) {
            argumentSize++;
            indexValid[1] = true;
        }
        if (pojoMethodParameters.isFieldIndex()) {
            argumentSize++;
            indexValid[2] = true;
        }
        if (pojoMethodParameters.isAnnotationIndex()) {
            argumentSize++;
            indexValid[3] = true;
        }
        if (pojoMethodParameters.isAnnotationTypeIndex()) {
            argumentSize++;
            indexValid[4] = true;
        }
        if (pojoMethodParameters.isMethodNameIndex()) {
            argumentSize++;
            indexValid[5] = true;
        }
        if (pojoMethodParameters.isFieldNameIndex()) {
            argumentSize++;
            indexValid[6] = true;
        }
        if (pojoMethodParameters.isTargetClassNameIndex()) {
            argumentSize++;
            indexValid[7] = true;
        }
        if (pojoMethodParameters.isTargetClassIndex()) {
            argumentSize++;
            indexValid[8] = true;
        }
        if (pojoMethodParameters.isAnnsorContext()) {
            argumentSize++;
            indexValid[9] = true;
        }
    }

    /**
     * Invokes the method that was decorated as being the Processor for registered Annotation.
     * @param element The AnnotationElement that includes all of the information about the Annotation.
     * @param handler The handler that will have its Processor method invoked.
     * @return Returns the value of the method invoked.
     * @throws IllegalAccessException    Thrown if there is a problem invoking the processor method.
     * @throws InvocationTargetException Thrown if there is a problem invoking the processor method.
     */
    public Object invoke(AnnotationElement element, Object handler) throws IllegalAccessException, InvocationTargetException {
        Object[] args = getMethodArguments(element, handler);
        return processorMethod.invoke(handler, args);
    }

    /**
     * This method creates the arguments that will be passed into the Processor method. It also will invoked
     * any convertor methods on the handler if they're provided.
     * @param element The AnnotationElement that includes all of the information about the Annotation.
     * @param handler The handler that will have its Processor method invoked.
     * @return The arguments to be passed into the Processor processorMethod.
     * @throws IllegalAccessException    Thrown if there is a problem invoking the convertor method.
     * @throws InvocationTargetException Thrown if there is a problem invoking the convertor method.
     */
    private Object[] getMethodArguments(AnnotationElement element, Object handler) throws IllegalAccessException, InvocationTargetException {

        Object[] args;

        if (convertable) {
            if (convertableMethod == null) {
                args = new Object[]{((Convertable) handler).convert(element)};
            } else {
                args = getMethodArguments(element, convertorParameterTypes);
                args = new Object[]{convertableMethod.invoke(handler, args)};
            }
        } else if (requiresMethodInspection()) {
            args = getMethodArguments(element, processorParameterTypes);
        } else {
            args = new Object[argumentSize];

            if (pojoMethodParameters.isTargetObjectIndex()) {
                args[pojoMethodParameters.getTargetObjectIndex()] = element.getTarget();
            }

            if (pojoMethodParameters.isMethodIndex()) {
                args[pojoMethodParameters.getMethodIndex()] = element.getMethod();
            }

            if (pojoMethodParameters.isFieldIndex()) {
                args[pojoMethodParameters.getFieldIndex()] = element.getField();
            }

            if (pojoMethodParameters.isAnnotationIndex()) {
                args[pojoMethodParameters.getAnnotationIndex()] = element.getAnnotation();
            }

            if (pojoMethodParameters.isAnnotationTypeIndex()) {
                args[pojoMethodParameters.getAnnotationTypeIndex()] = element.getScope();
            }

            if (pojoMethodParameters.isMethodNameIndex()) {
                args[pojoMethodParameters.getMethodNameIndex()] = element.getMethodName();
            }

            if (pojoMethodParameters.isFieldNameIndex()) {
                args[pojoMethodParameters.getFieldNameIndex()] = element.getFieldName();
            }

            if (pojoMethodParameters.isTargetClassNameIndex()) {
                args[pojoMethodParameters.getTargetClassNameIndex()] = element.getTargetClassName();
            }

            if (pojoMethodParameters.isTargetClassIndex()) {
                args[pojoMethodParameters.getTargetClassIndex()] = element.getTargetClass();
            }

            if (pojoMethodParameters.isMethodParametersIndex()) {
                args[pojoMethodParameters.getMethodParametersIndex()] = element.getMethodParameters();
            }

            if (pojoMethodParameters.isAnnsorContext()) {
                args[pojoMethodParameters.getAnnsorContextIndex()] = element.getContext();
            }
        }

        return args;
    }

    /**
     * This method inspects the provided method and creates a parameter list based upon the method
     * signature.
     * @param method The method to inspect and create the parameter list for.
     * @return The argument list that will be passed into the supplied method.
     */
    protected MethodArgumentType[] getMethodParameterTypes(Method method) {

        final Class[] parameterTypes = method.getParameterTypes();

        final MethodArgumentType[] methodParameters = new MethodArgumentType[parameterTypes.length];

        boolean hasField = false;

        boolean hasMethod = false;

        boolean hasConstructor = false;

        List<Integer> strings = new LinkedList<Integer>();

        int booleans = 0;

        for (int index = 0; index < parameterTypes.length; index++) {
            Class type = parameterTypes[index];
            if (type == Field.class) {
                methodParameters[index] = FIELD;
                hasField = true;
                booleans += 1;
            } else if (type == Method.class) {
                methodParameters[index] = METHOD;
                hasMethod = true;
                booleans += 1;
            } else if (type == Constructor.class) {
                methodParameters[index] = CONSTRUCTOR;
                hasConstructor = true;
                booleans += 1;
            } else if (type == Annotation.class || type.isAnnotation()) {
                methodParameters[index] = ANNOTATION;
                this.annotation = type;
            } else if (type == Class.class) {
                methodParameters[index] = TARGETCLASS;
            } else if (type == Scope.class) {
                methodParameters[index] = ANNOTATIONTYPE;
            } else if (AnnsorContext.class.isAssignableFrom(type)) {
                methodParameters[index] = ANNSORCONTEXT;
            } else if (type == AnnotationElement.class) {
                methodParameters[index] = ANNOTATIONELEMENT;
            } else if (type == Class[].class) {
                methodParameters[index] = PARAMETERS;
            } else if (type == String.class) {
                strings.add(index);
            } else {
                methodParameters[index] = TARGET;
            }
        }

        if (strings.size() > 0) {
            if (booleans == 1) { // Only one attribute type is passed in.
                if (hasField) {
                    methodParameters[strings.get(0)] = FIELDNAME;
                } else if (hasMethod) {
                    methodParameters[strings.get(0)] = METHODNAME;
                } else if (hasConstructor) {
                    methodParameters[strings.get(0)] = CONSTRUCTORNAME;
                }
            } else {
                for (Integer i : strings) {
                    methodParameters[i] = NAME;
                }
            }
        }

        return methodParameters;

    }

    /**
     * This method generates an Object[] of arguments that will be passed into the Processor method.
     * @param element          The AnnotationElement to extract the values from.
     * @param methodParameters The parameter type index.
     * @return An Object[] that will be passed into the method
     */
    private Object[] getMethodArguments(final AnnotationElement element, final MethodArgumentType[] methodParameters) {

        Object[] args = new Object[methodParameters.length];

        for (int index = 0; index < methodParameters.length; index++) {
            switch (methodParameters[index]) {
                case FIELD:
                    args[index] = element.getField();
                    break;
                case METHOD:
                    args[index] = element.getMethod();
                    break;
                case ANNOTATION:
                    Class annotationClass = parameterClassTypes[index];
                    if(annotationClass.isAssignableFrom(element.getAnnotation().getClass())){
                        args[index] = element.getAnnotation();
                    }else{
                        args[index] = null;
                    }
                    break;
                case ANNOTATIONELEMENT:
                    args[index] = element;
                    break;
                case ANNOTATIONTYPE:
                    args[index] = element.getScope();
                    break;
                case FIELDNAME:
                    args[index] = element.getFieldName();
                    break;
                case METHODNAME:
                    args[index] = element.getMethodName();
                    break;
                case CONSTRUCTOR:
                    args[index] = element.getConstructor();
                    break;
                case CONSTRUCTORNAME:
                    args[index] = element.getConstructor().getName();
                    break;
                case ANNSORCONTEXT:
                    args[index] = element.getContext();
                    break;
                case PARAMETERS:
                    args[index] = element.getMethodParameters();
                    break;
                case NAME:
                    if (element.getField() != null) {
                        args[index] = element.getFieldName();
                    } else if (element.getMethod() != null) {
                        args[index] = element.getMethodName();
                    } else if (element.getClass() != null) {
                        args[index] = element.getClass().getName();
                    } else if (element.getConstructor() != null) {
                        args[index] = element.getConstructor().getName();
                    }
                    break;
                case TARGET:
                    args[index] = element.getTarget();
                    break;
                case TARGETCLASS:
                    args[index] = element.getTargetClass();
                    break;
            }
        }
        return args;
    }

    /**
     * Returns true if the method should be inspected for the parameter list.
     * @return true if the method should be inspected, false otherwise.
     */
    boolean requiresMethodInspection() {
        for (boolean anIndexValid : indexValid) {
            if (!anIndexValid)
                return true;
        }
        return false;
    }

    protected Scope[] getScope() {
        return pojoMethodParameters.getScope();
    }


    public boolean acceptsAnnotation(Annotation annotation){
        return this.annotation != null && this.annotation.isAssignableFrom(annotation.getClass());
    }

    public boolean isAnnotationArgumentPresent(){
        return this.annotation != null;
    }

}