package net.java.dev.annsor.processors;

import net.java.dev.annsor.AnnotationElement;
import net.java.dev.annsor.Scope;
import net.java.dev.annsor.StatefulProcessor;
import net.java.dev.annsor.annotations.Destroy;
import net.java.dev.annsor.annotations.Processor;

import java.lang.reflect.Method;
import java.util.LinkedList;
import java.util.List;

/**
 * This implementation turns a POJO into a Annotation awared bean.
 * @author Richard L. Burton III
 */
public final class PojoProcessor implements StatefulProcessor {

    /**
     * A map of methods which can be invoked dynamically based upon their signature.
     */
    private final List<PojoMethod> processorList = new LinkedList<PojoMethod>();

    private PojoMethod destroyMethod;

    private final Object handler;

    /**
     * Constructs a wrapper around an object that was decorated with the @Processor annotation.
     * @param handler The target object that was decorated.
     */
    public PojoProcessor(Object handler) {
        Class clazz = handler.getClass();

        for (Method method : clazz.getDeclaredMethods()) {
            if (method.isAnnotationPresent(Processor.class)) {
                Processor annotation = method.getAnnotation(Processor.class);
                PojoMethod pmd = new PojoMethod(annotation, handler, method);
                processorList.add(pmd);
            }

            if (method.isAnnotationPresent(Destroy.class)) {
                Destroy annotation = method.getAnnotation(Destroy.class);
                destroyMethod = new PojoMethod(annotation, handler, method);
            }
        }
        this.handler = handler;
    }

    public boolean isStateful() {
        return destroyMethod != null;
    }

    /**
     * This method invokes all of the methods that were decorated with @Processor
     * annotations.
     * @param annotationElement An instance of the AnnotationElement thats being processed.
     * @throws Exception Thrown if there are any issues invoking the process method(s).
     */
    public void process(AnnotationElement annotationElement) throws Exception {
        for (PojoMethod method : processorList) {
            if (isInScope(method.getScope(), annotationElement.getScope())) {
                if (method.isAnnotationArgumentPresent()) {
                    if (method.acceptsAnnotation(annotationElement.getAnnotation())) {
                        method.invoke(annotationElement, handler);
                    }
                } else {
                    method.invoke(annotationElement, handler);
                }
            }
        }
    }

    /**
     * Checks to see if the Annotation is located in the scope that the @Processor method expects.
     * @param scopes The scopes the Processor method excepts.
     * @param scope  The scope to compare with to see if the process method should be invoked.
     * @return True if the processor method should be invoked.
     */
    private boolean isInScope(Scope[] scopes, Scope scope) {
        for (Scope s : scopes) {
            if (s.equals(Scope.ANY) || s.equals(scope)) {
                return true;
            }
        }
        return false;
    }

    public void destroy(AnnotationElement annotationElement) throws Exception {
        if (destroyMethod != null)
            destroyMethod.invoke(annotationElement, handler);
    }

}
