package net.java.dev.annsor.processors;

import net.java.dev.annsor.AnnotationElement;
import net.java.dev.annsor.ProcessorFactory;
import net.java.dev.annsor.annotations.Create;

import java.lang.reflect.Method;

/**
 * This custom ProcessorFactory is designed to inspect Pojo's
 * that have been decorated with @Factory and @Create and turn them into
 * a custom ProcessorFactory.
 * @author Richard L. Burton III
 * @version 1.0
 */
public class PojoProcessorFactory implements ProcessorFactory {

    /**
     * The <tt>create</tt> method to be invoked in order to create a new
     * custom Processor object.
     */
    protected PojoMethod createMethod;

    /**
     * The object object that has the create method.
     */
    protected Object target;

    /**
     * Constructs a new PojoProcessorFactory from a Pojo that was decorated with the @Factory
     * and @Create annotations.
     * @param target The object that will be used for the Factory.
     * @throws IllegalArgumentException Thrown if there was no @Create annotation on the
     *                                  factory object.
     */
    public PojoProcessorFactory(Object target) {
        this.target = target;
        for (Method m : target.getClass().getDeclaredMethods()) {
            inspectMethod(m, target);
        }

        if (createMethod == null) {
            throw new IllegalArgumentException("The provided @Factory was not decorated with a @Create method.");
        }
    }

    /**
     * This method is invoked by AnnoConainer when a registered annotation is found.
     * @param annotationElement The AnnotationElement that is being processed.
     * @return A new instance of an Processor object.
     * @throws Exception Thrown if there is a problem invoking the method.
     */
    public Object create(AnnotationElement annotationElement) throws Exception {
        return createMethod.invoke(annotationElement, target);
    }

    protected void inspectMethod(Method method, Object target) {
        if (method.isAnnotationPresent(Create.class)) {
            if (createMethod != null) {
                throw new IllegalArgumentException("Annsor already found an existing @Create method on the class " + target.getClass());
            }
            if (method.getReturnType() != void.class) {
                createMethod = new PojoMethod(method.getAnnotation(Create.class), target, method);
            } else {
                throw new IllegalArgumentException("The provided @Create method does not return any values.");
            }
        }

    }

}
