package net.java.dev.annsor.util;

import java.lang.annotation.Annotation;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.Field;
import java.lang.reflect.Method;

/**
 * @author Richard L. Burton III
 * @version 1.0
 */
public class AnnotationInspector {

    /**
     * Checks to see if the given annotation is present at the method level.
     * @param clazz      The class to inspect.
     * @param annotation The annotation in question.
     * @return true if it's found false otherwise.
     */
    public static boolean isPresentOnMethod(Class clazz, Class<? extends Annotation> annotation) {
        if (!isRetainedAtRuntime(annotation))
            return false;

        Method[] methods = clazz.getDeclaredMethods();
        for (Method method : methods) {
            if (method.isAnnotationPresent(annotation)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Checks to see if the given annotation is present at the class level.
     * @param clazz      The class to inspect.
     * @param annotation The annotation in question.
     * @return true if it's found false otherwise.
     */
    public static boolean isPresentOnClass(Class clazz, Class<? extends Annotation> annotation) {
        return isRetainedAtRuntime(annotation) && clazz.isAnnotationPresent(annotation);
    }

    /**
     * Checks to see if the given annotation is present at the field level.
     * @param clazz      The class to inspect.
     * @param annotation The annotation in question.
     * @return true if it's found false otherwise.
     */
    public static boolean isPresentOnField(Class clazz, Class<? extends Annotation> annotation) {
        if (!isRetainedAtRuntime(annotation))
            return false;

        Field[] fields = clazz.getDeclaredFields();
        for (Field field : fields) {
            if (field.isAnnotationPresent(annotation)) {
                return true;
            }
        }
        return false;
    }

    /**
     * checks to see if an annotation is rentained at runtime.
     * @param annotation The annotation in question.
     * @return true if the annotation rention policy is set to RUNTIME
     */
    public static boolean isRetainedAtRuntime(Class<? extends Annotation> annotation) {
        Retention retention = annotation.getAnnotation(Retention.class);
        return RetentionPolicy.RUNTIME.equals(retention.value());
    }

}
